"""Shampines Rosenbrock method for testing. To be generalized and complemented
by Rodas4, Rodas3 and 3rd and 4th order methods of Rank."""


class Sh2R(OdeSolver):
    order = 2
    order_secondary = 3

    d = 1/(2 + sqrt(2))
    e32 = 6 + sqrt(2)

    sc_params = "W"
    max_factor = MAX_FACTOR0          # initially
    min_factor = MIN_FACTOR

    def _init_min_step_parameters(self):
        """Define the parameters h_min_a and h_min_b for the min_step rule:
            min_step = max(h_min_a * abs(t), h_min_b)
        from RKSuite.
        """

        # minimum difference between distinct C-values
        cdiff = 0.5

        # determine min_step parameters
        epsneg = np.finfo(self.y.dtype).epsneg
        tiny = np.finfo(self.y.dtype).tiny
        h_min_a = 10 * epsneg / cdiff
        h_min_b = sqrt(tiny)
        return h_min_a, h_min_b

    def _init_sc_control(self, sc_params):
        # implement different controller, see Kennedy 2019
        coefs = {"G": (0.7, -0.4, 0, 0.9),      # Gustafsson
                 "S": (0.6, -0.2, 0, 0.9),      # Soderlind
                 "W": (2., -1., -1., 0.8),      # Watts
                 "standard": (1, 0, 0, 0.9)}
        # use default controller of method if not specified otherwise
        sc_params = sc_params or self.sc_params
        if (isinstance(sc_params, str) and sc_params in coefs):
            kb1, kb2, a, g = coefs[sc_params]
        elif isinstance(sc_params, tuple) and len(sc_params) == 4:
            kb1, kb2, a, g = sc_params
        else:
            raise ValueError('sc_params should be a tuple of length 4 or one '
                             'of the strings "G", "S", "W" or "standard"')
        # set all parameters
        self.minbeta1 = kb1 * self.error_exponent
        self.minbeta2 = kb2 * self.error_exponent
        self.minalpha = -a
        self.safety = g
        self.safety_sc = g ** (kb1 + kb2)
        self.standard_sc = True                                # for first step

    def _validate_jac(self, jac, sparsity):
        t0 = self.t
        y0 = self.y
        if jac is None:
            if sparsity is not None:
                if issparse(sparsity):
                    sparsity = csc_matrix(sparsity)
                groups = group_columns(sparsity)
                sparsity = (sparsity, groups)

            def jac_wrapped(t, y):
                self.njev += 1
                f = self.fun_single(t, y)
                J, self.jac_factor = num_jac(self.fun_vectorized, t, y, f,
                                             self.atol, self.jac_factor,
                                             sparsity)
                return J
            J = jac_wrapped(t0, y0)
        elif callable(jac):
            J = jac(t0, y0)
            self.njev += 1
            if issparse(J):
                J = csc_matrix(J, dtype=y0.dtype)

                def jac_wrapped(t, y):
                    self.njev += 1
                    return csc_matrix(jac(t, y), dtype=y0.dtype)
            else:
                J = np.asarray(J, dtype=y0.dtype)

                def jac_wrapped(t, y):
                    self.njev += 1
                    return np.asarray(jac(t, y), dtype=y0.dtype)

            if J.shape != (self.n, self.n):
                raise ValueError(
                    f"`jac` is expected to have shape {(self.n, self.n)},"
                    f" but actually has {J.shape}.")
        else:
            if issparse(jac):
                J = csc_matrix(jac, dtype=y0.dtype)
            else:
                J = np.asarray(jac, dtype=y0.dtype)
            if J.shape != (self.n, self.n):
                raise ValueError(
                    f"`jac` is expected to have shape {(self.n, self.n)},"
                    f" but actually has {J.shape}.")
            jac_wrapped = None
        return jac_wrapped, J

    def _validate_dfdt(self, dfdt):
        t0 = self.t
        y0 = self.y
        if dfdt is None:
            dfdt_wrapped = self._num_dfdt
            T = dfdt_wrapped(t0, y0)
            assert T.shape == (self.n, ), "numerical dtdf has wrong shape"
        elif callable(dfdt):
            def dfdt_wrapped(t, y):
                return np.asarray(dfdt(t, y), dtype=y0.dtype)

            T = dfdt_wrapped(t0, y0)
            if T.shape != (self.n, ):
                raise ValueError(
                    f"`dfdt` is expected to have shape {(self.n, )},"
                    f" but actually has {T.shape}.")
        else:
            T = np.asarray(dfdt, dtype=y0.dtype)
            if T.shape != (self.n, ):
                raise ValueError(
                    f"`dfdt` is expected to have shape {(self.n, )},"
                    f" but actually has {T.shape}.")
            dfdt_wrapped = None
        return dfdt_wrapped, T

    def _num_dfdt(self, t, y):
        """numerical approximation of dfdt by simple forward finite difference
        uses self.h_abs for scale
        """
        EPS = np.finfo(self.y.dtype).eps
        factor = sqrt(EPS)
        h = self.h_abs * self.direction
        dt = factor*h
        f_pert = self.fun_single(t + dt, y)
        df = f_pert - self.f
        return df/dt

    def _set_lu_functions(self, J):
        if issparse(J):
            def lu(A):
                self.nlu += 1
                return splu(A)

            def solve_lu(LU, b):
                NLS[()] += 1
                return LU.solve(b)
            I = eye(self.n, format='csc', dtype=self.y.dtype)
        else:
            def lu(A):
                self.nlu += 1
                return lu_factor(A, overwrite_a=True)

            def solve_lu(LU, b):
                NLS[()] += 1
                return lu_solve(LU, b, overwrite_b=True)
            I = np.identity(self.n, dtype=self.y.dtype)
        return lu, solve_lu, I

    def __init__(self, fun, t0, y0, t_bound, max_step=np.inf, rtol=1e-3,
                 atol=1e-6, jac=None, jac_sparsity=None, dfdt=None,
                 vectorized=False, first_step=None, sc_params=None,
                 **extraneous):
        warn_extraneous(extraneous)
        super().__init__(fun, t0, y0, t_bound, vectorized,
                         support_complex=True)
        self.max_step = validate_max_step(max_step)
        self.rtol, self.atol = validate_tol(rtol, atol, self.y)
        self.f = self.fun(self.t, self.y)
        if self.f.dtype != self.y.dtype:
            raise TypeError('dtypes of solution and derivative do not match')
        self.h_min_a, self.h_min_b = self._init_min_step_parameters()
        self.tiny_err = self.y.size**0.5 * np.finfo(self.y.dtype).epsneg**0.8
        order_error = min(self.order_secondary, self.order)
        self.error_exponent = -1 / (order_error + 1)
        self._init_sc_control(sc_params)

        self.h_previous = None
        self.y_old = None                              # value on diagonal
        NFS[()] = 0                                # global failed step counter
        NLS[()] = 0                       # global linear system solves counter

        # jac matters
        # J is considered current if it is determined at the start of the
        # current step, not at (t_stage, y_predict). I chooose to use the same
        # J throughout the step. This has the disadvantage of lost work if the
        # last stage fails. The advantage is reduced complexity and quicker
        # stepsize reduction.
        self.current_J = True
        self.jac_factor = None      # (for numerical jac approximation)
        self.jac, self.J = self._validate_jac(jac, jac_sparsity)
        self.linear = self.jac is None
        self.lu, self.solve_lu, self.M = self._set_lu_functions(self.J)
        self.LU = None
        self.h_LU = None

        # size of first step:
        if first_step is None:
            b = self.t + self.direction * min(
                abs(self.t_bound - self.t), self.max_step)
            J = self.J
            if issparse(J):
                J = J.todense()
            self.h_abs = abs(h_start(
                self.fun, self.t, b, self.y, self.f,
                order, self.rtol, self.atol, J=J))
        else:
            self.h_abs = validate_first_step(first_step, t0, t_bound)

        # time derivative of fun, dfdt:
        self.dfdt, self.T = self._validate_dfdt(dfdt)

    def _step_impl(self):
        t = self.t
        y = self.y
        h_abs, min_step = self._reassess_stepsize(t, self.h_abs)

        F0 = self.f
        T = self.T
        J = self.J

        # loop until the step is accepted
        step_accepted = False
        step_rejected = False
        while not step_accepted:
            if h_abs < min_step:
                return False, self.TOO_SMALL_STEP
            h = h_abs * self.direction
            t_new = t + h

            hd = h*self.d
            W = self.M - hd*J
            LU = self.lu(W)                                         # factorize

            k1 = self.solve_lu(LU, F0 + hd*T)                       # solve
            F1 = self.fun(t + 0.5*h, y + 0.5*h*k1)                  # eval
            k2 = self.solve_lu(LU, F1 - k1) + k1                    # solve
            y_new = y + h*k2

            # FSAL for err
            F2 = self.fun(t_new, y_new)                             # eval
            k3 = self.solve_lu(LU,                                  # solve
                               F2 - self.e32*(k2 - F1) - 2*(k1 - F0) + hd*T)
            err = h/6*(k1 - 2*k2 + k3)
            # The secondary method is only A-stable. Yet, err is NOT filtered!
            # This method should be more reliable without filtering.

            # calculate error norm
            scale = calculate_scale(self.atol, self.rtol, y, y_new)
            error_norm = norm(err/scale)
            if not isfinite(error_norm):                              # useful?
                warn(f"Integration error norm is not finite. ({t = })")

            # propose new stepsize
            step_accepted, h_abs = self._assess_error_and_stepsize(
                error_norm, h_abs, step_rejected)
            step_rejected = not step_accepted

        # set or store for next step or interpolant
        self.y_old = y.copy()
        self.f_old = self.f.copy()
        self.f = F2.copy()
        self.error_norm_old = error_norm
        self.h_previous = h
        self.h_abs = h_abs
        self.k1 = k1
        self.k2 = k2

        # output
        self.t = t_new
        self.y = y_new

        # step accepted
        # update J and dfdt for next step
        if t_new != self.t_bound:
            if self.jac is not None:
                self.J = self.jac(t_new, y_new)
            if self.dfdt is not None:
                self.T = self.dfdt(t_new, y_new)

        return True, None

    def _assess_error_and_stepsize(self, error_norm, h_abs, step_rejected):
        # evaluate error
        if error_norm < 1:
            step_accepted = True
            if error_norm < self.tiny_err:
                factor = self.max_factor
                self.standard_sc = True
            elif self.standard_sc:
                factor = self.safety * error_norm**self.error_exponent
                self.standard_sc = False
            else:
                # use second order controller
                h_ratio = h_abs*self.direction/self.h_previous
                factor = self.safety_sc*(
                        error_norm**self.minbeta1 *
                        self.error_norm_old**self.minbeta2 *
                        h_ratio**self.minalpha)
                # limit max step increase depending on matrix norm:
                # to 1.2 for very stiff problems.
                # This is an extra term from Shampine, makes sense for
                # Rosenbrock methods that have a single iteration with a J from
                # the start of the step.
                Jnorm = self._norm_M1(self.J)
                factor = min(1.2 + (self.max_factor-1.2)/(1+Jnorm/50),
                             max(self.min_factor, factor))
            if step_rejected:
                factor = min(1, factor)
        else:
            step_accepted = False
            NFS[()] += 1
            factor = max(self.min_factor,
                         self.safety * error_norm**self.error_exponent)

        # reduce max_factor when on scale
        if factor < MAX_FACTOR:
            self.max_factor = MAX_FACTOR

        # new step size
        h_abs *= factor
        return step_accepted, h_abs

    def _reassess_stepsize(self, t, h_abs):
        # limit step size
        min_step = max(self.h_min_a * (abs(t) + h_abs), self.h_min_b)
        if h_abs < min_step or h_abs > self.max_step:
            h_abs = min(self.max_step, max(min_step, h_abs))
            self.standard_sc = True

        # handle final integration steps
        d = abs(self.t_bound - t)                          # remaining interval
        if d < 2 * h_abs:
            if d > h_abs:
                # h_abs < d < 2 * h_abs: "look ahead".
                # split d over last two steps. This reduces the chance of a
                # very small last step.
                h_abs = max(0.5 * d, min_step)
                self.standard_sc = True
            else:
                # d <= h_abs: Don't step over t_bound
                h_abs = d

        return h_abs, min_step

    def _norm_M1(self, M):
        """1-norm of matrix"""
        # maximum column sum of abs values.
        return np.abs(M).sum(axis=0).max()

    def _dense_output_impl(self):
        return Sh2R_interpolant(self.t_old, self.t, self.y_old,
                                self.k1, self.k2)


class Sh2R_interpolant(DenseOutput):
    """quadratic interpolant for Sh2R
    """
    def __init__(self, t_old, t, y_old, k1, k2):
        super().__init__(t_old, t)
        self.h = t - t_old
        self.y_old = y_old
        self.k1 = k1
        self.k2 = k2

    def _call_impl(self, t, d=Sh2R.d):

        # scaled time
        s = (t - self.t_old) / self.h

        # output
        y = self.y_old[:, np.newaxis] + self.h*s/(1 - 2*d)*(
                (1 - s)*self.k1[:, np.newaxis] +
                (s - 2*d)*self.k2[:, np.newaxis])

        if t.shape:
            return y
        else:
            return y[:, 0]

